#/* finterp.c
**
** scan a message using a filter
**
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "prefs.h"
#include "mrmacs.h"
#include "inline-funcs.h"
#include "regex/regex.h"
#include "tools.h"
#include "vxerror.h"
#include "hlsock.h"
#include "tmpfile.h"

#include "userdb.h"
#include "rexec.h"
#include "smtp.h"
#include "deliver.h"

#include "hread.h"
#include "fparser.h"
#include "finterp.h"
#include "emHeader.h"

/* Forwrts-Deklarationen fr die von `finterp' aus aufgerufenen Funktionen. */
static int fi_ccheck (condlist_t cond, char *line);
static int fi_andcheck (condlist_t cond, char *line);
static int fi_action (action_t action, int pre, FILE *mfp,
		      int *acerror, xFile tgt, int err);
void fi_error (int acerror);

/* Diese Routine mu mit `1' zurckkehren, wenn die E-Mail nach ihrer
** Abarbeitung noch vorhanden ist und von `vxdeliver' normal weiterverarbeitet
** werden soll.
*/
int finterp (filter_t filter, FILE *mfp, xFile tgt, int err)
{
    filter_t fi;
    char *li;
    int acerror;
    /* Dieses Flag zeigt den aktuellen Teil der E-Mail an, der gerade
    ** bearbeitet wird.
    */
    int head = 1, acres;
    /* Im ersten Schritt wird die E-Mail zeilenweise eingelesen, wobei
    ** zunchst jede eingelesene Kopfzeile durch die Bedingungen smtlicher
    ** Filter gejagt wird, die als Kopffilter markiert sind und anschlieend
    ** alle Zeilen des `message body' durch die als Rumpffilter markierten
    ** Filter geschickt.
    */
    while (noNULL(li = hread_line (mfp, head))) {
	if (*li == '\0') { head = 0; continue; }
	for (fi = filter; noNULL(fi); fi = fi->next) {
	    if (fi->result) { continue; }
	    if ((head && !fi->head) || (!head && !fi->body)) { continue; }
	    if (fi_ccheck (fi->conditions, li)) { fi->result = 1; }
	}
    }
    /* Die Aktionen mssen wiederum die gesamte E-Mail lesen - darum mu der
    ** Dateizeiger zurrkgesetzt werden.
    */
    rewind (mfp);
    for (fi = filter; noNULL(fi); fi = fi->next) {
	/* Die Aktionen eines Filters, dessen Bedingungen nicht zutreffen
	** werden selbstverstndlich nicht ausgefhrt.
	*/
	if (!fi->result) { continue; }
	/* Wenn `fi_action' `0' zurckgibt, dann heit das, da die letzte
	** ausgefhrte Aktion entweder `Drop' oder ein `Mailbox' war oder
	** eine `Exec'-Aktion einen Exit-Status von 2 hatte. In diesem Fall
	** ist die Arbeit von `VxDeliver' beendet.
	*/
	acres = fi_action (fi->action, fi->pre, mfp, &acerror, tgt, err);
	/* acres == 0 bedeutet, da die Arbeit von `Vxeliver' abgeschlossen
	** ist.
	*/
	if (!acres) { return 0; }
	/* acres == -1 bedeutet, da eine `Exec'-Aktion einen Fehler
	** erzeugt hat.
	*/
	if (acres == -1) { fi_error (acerror); return 0; }
	/* Jedes andere Ergebnis bewirkt die Fortsetzung der Arbeit. */
    }
    return 1;
}

void fi_error (int acerror)
{
    ;
}

static int fi_ccheck (condlist_t cond, char *line)
{
    condlist_t ol;
    regmatch_t pmatch[1];
    for (ol = cond; noNULL(ol); ol = ol->next) {
	ifnull (ol->also) {
	    if (!regexec (&ol->rx, line, 1, pmatch, 0)) { return 1; }
	} else {
	    if (fi_andcheck (ol, line)) { return 1; }
	}
    }
    return 0;
}

static int fi_andcheck (condlist_t cond, char *line)
{
    condlist_t al;
    regmatch_t pmatch[1];
    int af = 1;
    /* Bei einem `UND' mssen smtliche Bedingungen erfllt sein, damit
    ** die Gesamtbedingung als erfllt zu betrachten ist.
    */
    for (al = cond; noNULL(al); al = al->also) {
	/* Allerdings knnen einige der Teilbedingungen durch vorherige
	** Lufe bereits erfllt worden sein.
	*/
	if (al->alsores) { continue; }
	/* Ansonsten wird die Bedingung berprft. */
	if (regexec (&al->rx, line, 1, pmatch, 0)) {
	    /* Trifft sie nicht zu, dann ist das Gesamtergebnis in jedem
	    ** Falle falsch. Allerdings werden trotzdem alle noch folgenden
	    ** Bedingungen berprft; es knnte sich ja eine von ihnen als
	    ** erfllt erweisen, die dann - sobald diese Bedingung hier
	    ** (in einem weiteren Lauf) erfllt ist - zusammen mit dieser
	    ** ein positives Resultat ergibt.
	    */
	    af = 0;
	} else {
	    /* Wenn die Bedingung erfllt ist, dann mu das Flag `alsores'
	    ** in der Bedingungsstruktur gesetzt werden.
	    */
	    al->alsores = 1;
	}
    }
    return af;
}

static void do_Mailbox (ehList_t hl, FILE *body, xFile tgt, char *name);
static int do_MailTo (char *from, char *to, char *msg, int err);
static int do_ForwardTo (char *from, char *to, ehList_t hl, FILE *mfp, int err);
static int do_Exec (uid_t uid, gid_t gid, char *cmd,
		    ehList_t hl, FILE *mfp, FILE *new);

static emHeader_t get_headers (FILE *fp)
{
    emHeader_t hdr;
    char *li, *dp, *p;
    hdr.f = hdr.l = (ehList_t) NULL;
    rewind (fp);
    while (noNULL(li = hread_line (fp, 1))) {
	if (!*li) { break; }
	if (!strncmp (li, "From ", 5)) {
	    dp = &li[5];
	    emHeader_ins (&hdr, "From ", dp);
	} else if (noNULL(dp = strchr (li, ':'))) {
	    p = dp; *dp++ = '\0'; ++dp;
	    while (isws (*--p)) { *p = '\0'; }
	    while (isws (*li)) { ++li; }
	    emHeader_ins (&hdr, li, dp);
	}
    }
    return hdr;
}

/* fi_action fhrt eine Liste von Filteraktionen aus. Diese Funktion kann
** durch eines von drei Ereignissen beendet werden:
** a) Eine Mailbox-Aktion wurde durchgefhrt und die E-Mail in einer Mailbox
**    abgelegt;
** b) Eine Drop-Aktion wurde durchgefhrt und die E-Mail verworfen;
** c) Die Liste der Aktionen wurde abgearbeitet und die Mail-Datei durch
**    die genderte E-Mail ersetzt.
*/
static int fi_action (action_t action, int pre, FILE *_mfp,
		      int *acerror, xFile tgt, int err)
{
    action_t ac;
    FILE *nf, *mfp = (FILE *) NULL;
    uid_t uid; gid_t gid;
    fpos_t pos;
    char *saddr, *cmd;
    int ev;
    emHeader_t hdr = { (ehList_t) NULL, (ehList_t) NULL };
    if (fvxtrans (_mfp, &mfp)) {
	/* Fehlschlag auf der ganzen Linie! Die Maildatei konnte nicht
	** kopiert werden.
	*/
	return -99;
    }
    fgetpos (mfp, &pos);
    for (ac = action; noNULL(ac); ac = ac->next) {
	ifnull (hdr.f) {
	    hdr = get_headers (mfp);
	    fgetpos (mfp, &pos);
	}
	switch (ac->action) {
	    case ac_none: break;
	    case ac_insert:
		/* Smtliche E-Mail-Header einlesen - wenn sie nicht bereits
		** im Speicher stehen; dann den Header mit dem entsprechenden
		** Namen suchen (`case independant'); wenn der Header gefunden
		** wurde, den Wert anhngen; wurde der Header nicht gefunden,
		** dann mu er eingefgt werden.
		*/
		vxlog (2, "performing as Insert filter action.\n");
		emHeader_ins (&hdr, ac->a.insert.elem, ac->a.insert.new);
		break;
	    case ac_replace:
		/* Smtliche E-Mail-Header einlesen - wenn sie nicht bereits
		** im Speicher stehen; dann den Header mit dem entsprechenden
		** Namen suchen (`case independant'); danach den Wert des
		** Headers ersetzen.
		*/
		vxlog (2, "performing a Replace filter action.\n");
		emHeader_chg (&hdr, ac->a.replace.elem, ac->a.replace.ix,
				    ac->a.replace.new);
		break;
	    case ac_delete:
		/* Smtliche E-Mail-Header einlesen - wenn sie nicht bereits
		** im Speicher stehen; dann den Header mit dem entsprechenden
		** Namen suchen (`case independant') und lschen.
		*/
		vxlog (2, "performing a Delete filter action.\n");
		emHeader_del (&hdr, ac->a.delete.elem, ac->a.delete.ix);
		break;
	    case ac_mailto:
		/* SMTP-Verbindung zu `nonlocalto' aufbauen; E-Mail mittels
		** Protokoll versenden; die E-Mail liegt als eine lange
		** Zeichenkette vor. Es mu `byte stuffing' durchgefhrt
		** werden.
		*/
		vxlog (2, "performing a MailTo filter action.\n");
		ev = do_MailTo (ac->a.mailto.from, ac->a.mailto.to,
				ac->a.mailto.msg, err);
		break;
	    case ac_forward:
		/* SMTP-Verbindung zu `nonlocalto' aufbauen; bisherigen
		** Message-Text (d.h. zunchst die (evtl im Speicher stehenden)
		** Header ausgeben; danach eine Leerzeile ausgeben; danach den
		** Message-Rumpf ausgeben) versenden.
		*/
		vxlog (2, "performing a ForwardTo filter action.\n");
		fsetpos (mfp, &pos);
		ev = do_ForwardTo (ac->a.forward.from, ac->a.forward.to,
				   hdr.f, mfp, err);
		break;
	    case ac_drop:
		/* Rckgabe von `0'. */
		vxlog (2, "performing a Drop filter action.\n");
		fclose (mfp);
		emHeader_free (&hdr);
		return 0;
	    case ac_exec:
		/* Externen Mail-Filter erzeugen mit der aktuellen Message
		** als Standardeingabe und einer neuen temporren Datei als
		** Ausgabe. Wenn der externe Mail-Filter mit einem Exit-Code
		** == 0 zurckkehrt, dann die neu angelegte (Ausgabe-)Datei 
		** in die Mailbox legen; kehrt der externe Filter mit einem
		** Exit-Code != 0 zurck, dann die neu angelegte Datei
		** verwerfen.
		** Wenn der Exit-Status == 0 ist (ext. Filter hat den Versand
		** bernommen), dann Ausgabedatei verwerfen und weitermachen.
		** Wenn der Exit_Status == 1 ist, dann die Header im Speicher
		** lschen, und die ursprngliche Mail-Datei durch die neue
		** ersetzen.
		*/
		vxlog (2, "performing an Exec filter action.\n");
		vxlog (2, "");
		saddr = gconc (__FILE__, __LINE__, "VxMail@", mailhost());
		ifnull (cmd = valid_prog(ac->a.exec.cmd, &uid, &gid, 1)) {
		    vxlog (0, "Exec: no permission to execute `%s'.\r\n",
			      ac->a.exec.cmd);
		    deliver_error (ac->a.exec.from, saddr,
				   ac->a.exec.from, NULL,
				   "no permission to execute program\r\n",
				   err);
		    break;
		}
		ifnull (nf = fvxtmp()) {
		    vxlog (0, "Exec: failed to open temporary file - %s\n",
			      strerror(errno));
		    deliver_error (ac->a.exec.from, saddr,
				   ac->a.exec.from, NULL,
				   "creating temporary file failed.\r\n",
				   err);
		    break;
		}
		ev = do_Exec (uid, gid, cmd, hdr.f, mfp, nf);
		rewind (nf);
		gfree (cmd);
		if (ev == 1) {
		    /* Die ursprngliche Datei mu durch die neue ersetzt
		    ** und die Header-Liste weggeworfen werden.
		    */
		    fclose (mfp); mfp = nf; nf = NULL;
		    emHeader_free (&hdr);
		} else {
		    /* Es gibt zwei mgliche Flle:
		    ** a) Alle Arbeiten wurden durch den externen Filter
		    **    erledigt. Es gibt keine Ausgaben und die Eingabe
		    **    mu geschlossen werden. (ev == 0)
		    ** b) Bei der Verarbeitung der Mail durch den externen
		    **    Filter ist ein Fehler aufgetreten. Es gibe keine
		    **    Ausgaben und die Eingabe mu geschlossen werden.
		    **    (ev < 0 || ev > 1)
		    */
		    fclose (nf); fclose (mfp);
		    emHeader_free (&hdr);
		    return ev;
		}
		break;
	    case ac_mailbox:
		/* Mail in der (angegebenen) Mailbox ablegen (im Moment ist
		** das immer die `INBOX'); anschlieend Rckkehr mit 0.
		*/
		vxlog (2, "performing a Mailbox filter action.\n");
		do_Mailbox (hdr.f, mfp, tgt, "INBOX");
		/*do_Mailbox (hdr.f, mfp, tgt, ac->a-mailbox.name);*/
		fclose (mfp);
		emHeader_free (&hdr);
		return 0;
	}
    }
    /* Ersetze die Mail-Datei durch eine neue, die die vernderten Header
    ** sowie den (evtl. vernderten) Rumpf der E-Mail enthlt.
    */
    if (pre) { fvxtrans (mfp, &_mfp); }
    fclose (mfp);
    emHeader_free (&hdr);
    return 1;
}

static void do_Mailbox (ehList_t hl, FILE *body, xFile tgt, char *name)
{
    char line[1024];
    while (noNULL(hl)) {
	xPuts (hl->name, tgt);
	if (lccmp (hl->name, "From ", -1)) { xPuts (": ", tgt); }
	xPuts (hl->value, tgt);
	xPuts ("\r\n", tgt);
	hl = hl->next;
    }
    xPuts ("\r\n", tgt);
    while (noNULL(fgets (line, 1024, body))) {
	xPuts (line, tgt);
    }
    xSet (tgt);
}

/* open a mail connection... */
static HLSock open_mail (char *action, char *from, char *to, int err)
{
    char *msg = NULL, *msg2 = NULL, *s;
    HLSock hlc = (HLSock) NULL;
    int portnum;
    ifnull (prefs.nonlocalto) {
	vxlog (0, "%s: `nonlocalto' undefined.\n", action);
	msg = "Not configured for sending remote messages.\r\n"
	      "Please contact your system administrator!\r\n";
	goto mail_error;
    }
    portnum = LLPort ("smtp", "tcp");
    if (portnum < 0) {
	vxlog (0, "%s: service ==> portnum translation failed!\n", action);
	msg = "Initialising SMTP connection failed - unknown service.\r\n"
	      "Please contact your system administrator!\r\n";
	goto mail_error;
    }
    ifnull (hlc = HLConnect ("tcp", prefs.nonlocalto, portnum)) {
	vxlog (0, "%s: connecting host `%s' failed - %s.\n",
		  action, mailhost(), strerror (errno));
	msg = "Connecting host `$(HOST)' failed.\r\n";
	msg2 = mailhost();
	goto mail_error;
    }
    if (smtp_init (hlc, mailhost())) {
	vxlog (0, "%s: SMTP protocol error.\n", action);
	msg = "SMTP protocol error.\r\n";
	goto mail_error;
    }
    if (smtp_envelope (hlc, from, to)) {
	vxlog (0, "%s: SMTP protocol error.\n", action);
	msg = "SMTP protocol error.\r\n";
	goto mail_error;
    }
    if (smtp_bodyinit (hlc)) {
	vxlog (0, "MailTo: SMTP protocol error.\n");
	msg = "SMTP protocol error.\r\n";
	goto mail_error;
    }
    return hlc;
mail_error:
    unlessnull (hlc) { HLClose (hlc); }
    unlessnull (msg2) {
	s = gvarins (__FILE__, __LINE__, msg, "HOST", msg2);
    } else {
	s = gdupl (__FILE__, __LINE__, msg);
    }
    deliver_error (to, from, to, NULL, s, err);
    gfree (s);
    return (HLSock) NULL;
}

static char *add_domain (char *addr)
{
    ifnull (strrchr (addr, '@')) {
	return gconc (__FILE__, __LINE__, addr, "@", mailhost(), NULL);
    } else {
	return gdupl (__FILE__, __LINE__, addr);
    }
}

/* MailTo action... */
static int do_MailTo (char *_from, char *to, char *msg, int err)
{
    char *emsg, *saddr, *from = gdupl(__FILE__, __LINE__, _from);
    int stv;
    HLSock hlc = (HLSock) NULL;
    saddr = gconc (__FILE__, __LINE__, "VxMail@", mailhost(), NULL);
    /*from = add_domain (_from);*/
    hlc = open_mail ("MailTo", from, to, err);
    ifnull (hlc) { gfree (saddr); gfree (from); return 1; }
    HLPuts (msg, hlc); vxlog (5, "%s.\n", msg);
    HLPuts (".\r\n", hlc);
    stv = smtp_result (hlc);
    if (stv / 100 != 2) {
	vxlog (0, "MailTo: SMTP protocol error - message body failed.\n");
	emsg = "SMTP protocol error: sending message body failed.\r\n";
	goto MailTo_error;
    }
    HLPuts ("QUIT\r\n", hlc);
    stv = smtp_result (hlc);
    if (stv / 100 != 2) {
	vxlog (0, "MailTo: SMTP protocol warning - QUIT failed.\n");
    }
    HLClose (hlc); gfree (saddr); gfree (from);
    return 0;
MailTo_error:
    unlessnull (hlc) { HLClose (hlc); }
    deliver_error (from, saddr, from, NULL, emsg, err);
    gfree (saddr); gfree (from);
    return 2;
}

static int do_ForwardTo (char *_from, char *to, ehList_t hl,
			 FILE *mfp, int err)
{
    char *msg, *saddr, line[1024], *from;
    HLSock hlc = (HLSock) NULL;
    int waseol, stv;
    saddr = gconc (__FILE__, __LINE__, "VxMail@", mailhost(), NULL);
    from = add_domain (_from);
    hlc = open_mail ("ForwardTo", from, to, err);
    ifnull (hlc) { gfree (saddr); gfree (from); return 1; }
    while (noNULL(hl)) {
	if (strcmp(hl->name, "From ")) {
	    HLPrint (hlc, "%s: %s\r\n", hl->name, hl->value);
	    vxlog (5, "%s: %s\n", hl->name, hl->value);
	}
	hl = hl->next;
    }
    HLPuts ("\r\n", hlc); vxlog (5, "\n");
    while (noNULL(fgets (line, 1024, mfp))) {
	waseol = cuteol (line);
	HLPuts (line, hlc); vxlog (5, "%s", line);
	if (waseol) { HLPuts ("\r\n", hlc); vxlog (5, "\n"); }
    }
    HLPuts (".\r\n", hlc); vxlog (5, ".\n");
    stv = smtp_result (hlc);
    if (stv / 100 != 2) {
	vxlog (0, "ForwardTo: SMTP protocol error - message body failed.\n");
	msg = "SMTP protocol error: sending message body failed.\r\n";
	goto ForwardTo_error;
    }
    HLPuts ("QUIT\r\n", hlc); vxlog (5, "QUIT\n");
    stv = smtp_result (hlc);
    if (stv / 100 != 2) {
	vxlog (0, "ForwardTo: SMTP protocol warning - QUIT failed.\n");
    }
    HLClose (hlc); gfree (saddr); gfree (from);
    return 0;
ForwardTo_error:
    unlessnull (hlc) { HLClose (hlc); }
    deliver_error (from, saddr, from, NULL, msg, err);
    gfree (saddr); gfree (from);
    return 2;
}

static int do_Exec (uid_t uid, gid_t gid, char *cmd,
		    ehList_t hl, FILE *mfp, FILE *new)
{
    /* Was ist zu tun?
    ** a) Erzeugung einer pipe
    ** b) Starten eines Prozesses
    ** c) Ausgabe der Header mittels der pipe an den neuen Proze
    ** d) Ausgabe des Restes der Mail-Datei an den neuen Proze
    ** e) Schlieen der pipe
    ** f) Warten auf die Beendigung des Kind-Prozesses
    ** g) Rckkehr mit dem Exit-Status des Kind-Prozesses.
    */
    int childin[2], status;
    pid_t child, cst;
    char line[1024];
    ehList_t eh;
    if (pipe (childin)) { return -1; }
    switch ((child = fork())) {
	case -1: /* ERROR */
	    /* fork() fehlgeschlagen */
	    close (childin[0]); close (childin[1]);
	    return -2;
	case 0:  /* CHILD */
	    {   /* Hier mache ich eine Block auf, weil ich
		** eine lokale Variable bentige.
		*/
		extern char **environ;
		/* Argumentvektor fr den Shell-Aufruf */
		char *args[4];
		/* Flle den Argumentvektor: `/bin/sh' `-c' `cmd' NULL */
		args[0] = "/bin/sh";
		args[1] = "-c";
		args[2] = cmd;
		args[3] = NULL;
		/* Schliee den Ausgabedeskriptor der Pipe, weil der neue
		** Proze aus ihr lesen soll.
		*/
		close (childin[1]);
		/* Mache den Eingabedeskriptor der Pipe zur Standardeingabe des
		** Prozesses und schliee anschlieend den nun berflssigen
		** Filedeskriptor.
		*/
		if (dup2 (childin[0], 0) < 0) { exit (-1); }
		close (childin[0]);
		/* Mache den Filedeskriptor der (als Argument bergebenen)
		** Ausgabedatei zur Standardausgabe des Prozesses und
		** schliee anschlieend die (lokale Kopie der) Ausgabedatei.
		*/
		if (dup2 (fileno (new), 1) < 0) { exit (-1); }
		lseek (1, (off_t) 0, SEEK_SET);
		ftruncate (1, (off_t) 0);
		fclose (new);
		/* Setze user- und group-ID des Prozesses neu. */
		setreuid (getuid(), getuid());
		setgid (gid);
		setreuid (uid, uid);
		/* Und fhre anschlieend `/bin/sh -c "cmd"' aus. */
		execve (args[0], args, environ);
		/* Und fhre anschlieend `/bin/sh -c "cmd"' aus. */
	    }
	    exit (-2);
	default: /* PARENT */
	    /* Gib dem Kind-Proze eine Chance, zu terminieren, wenn in ihm
	    ** ein Fehler aufgetreten ist (bevor das Kommando gestartet werden
	    ** konnte.
	    */
	    sleep (1);
	    cst = waitpid (child, &status, WNOHANG);
	    if (cst > 0) {
		if (WIFEXITED (status)) {
		    return -3;
		} else if (WIFSIGNALED (status)) {
		    return -4;
		}
	    }
	    /* Der Elternproze bentigt dein Eingabedeskriptor des Pipe
	    ** nicht.
	    */
	    close (childin[0]);
	    /* Nun mu die modifizierte E-Mail an den Kind-Proze bergeben
	    ** werden - zunchst die Header,
	    */
	    eh = hl;
	    while (noNULL(eh)) {
		write (childin[1], eh->name, strlen (eh->name));
		if (lccmp (eh->name, "From ", -1)) {
		    write (childin[1], ": ", 2);
		}
		write (childin[1], eh->value, strlen (eh->value));
		write (childin[1], "\r\n", 2);
		eh = eh->next;
	    }
	    /* anschlieend der Rumpf der E-Mail. */
	    write (childin[1], "\r\n", 2);
	    while (noNULL(fgets (line, 1024, mfp))) {
		write (childin[1], line, strlen (line));
	    }
	    /* Signalisiere dem Kind-Proze das Ende der Daten durch schlieen
	    ** des Ausgabedeskriptors der Pipe.
	    */
	    close (childin[1]);
	    /* Nun warte regulr auf die Termination des Kind-Prozesses. */
	    waitpid (child, &status, 0);
	    if (WIFEXITED(status)) {
		/* Ein Exit-Code 1 entspricht einer normalen Termination
		** - mit dem Unterschied, da keine Ausgabedaten vorliegen.
		*/
		if (WEXITSTATUS(status) == 1) { return 1; }
		/* Jeder andere Exit-Code (mit Ausnahme von 0) entspricht
		** einem Verarbeitungsfehler im Kind-Proze.
		*/
		if (WEXITSTATUS(status)) { return -5; }
	    } else if (WIFSIGNALED(status)) {
		/* Wenn der Kind-Proze mit einem Signal terminiert wurde,
		** dann handelt es sich um eine Unterbrechung von auen
		** oder aber einen schweren Verarbeitungsfehler (SIGSEGV oder
		** SIGBUS) im Kind-Proze. In diesem Fall mu das Funktions-
		** ergebnis negativ sein.
		*/
		return -6;
	    }
    }
    /* Wenn die Ausfhrung von `do_Exec' an dieser Stelle angelangt ist,
    ** dann ist der Kind-Proze mit einem Exit-Code 0 zurckgekehrt. In
    ** diesem Fall ist auch das Funktionsergebnis von `do_Exec' 0.
    */
    return 0;
}
